import os

import numpy as np

g_nbytes_per_float = 8

def get_nstates(distn_name, trans_name):
    statinfo = os.stat(distn_name)
    nstates_distn, err_a = divmod(statinfo.st_size, g_nbytes_per_float)
    statinfo = os.stat(trans_name)
    nstates_squared, err_b = divmod(statinfo.st_size, g_nbytes_per_float)
    if err_a or err_b:
        raise ValueError('input file sizes should be multiples of 8 bytes')
    if nstates_distn*nstates_distn != nstates_squared:
        msg_a = 'the transition matrix should have the square '
        msg_b = 'of the number of elements in the initial distribution vector'
        raise ValueError(msg_a + msg_b)
    return nstates_distn

def load_distn_and_trans(distn_name, trans_name):
    with open(distn_name, 'rb') as fin:
        np_distn = np.fromfile(fin, dtype=float)
        nstates = len(np_distn)
    with open(trans_name, 'rb') as fin:
        np_trans = np.fromfile(fin, dtype=float).reshape((nstates, nstates))
    return np_distn, np_trans



"""A thin wrapper around the C extension.
In this module the term 'matrix' is abused to mean numpy array.
"""

import itertools

import numpy as np

import hmmusbuf

def is_stochastic_vector(v):
    if len(v.shape) != 1:
        return False
    if any(x<0 for x in v):
        return False
    if abs(1.0 - v.sum()) > 1e-7:
        return False
    return True

def is_square_matrix(M):
    if len(M.shape) != 2:
        return False
    if len(set(M.shape)) != 1:
        return False
    return True

def is_right_stochastic_matrix(M):
    if not is_square_matrix(M):
        return False
    if not all(is_stochastic_vector(v) for v in M):
        return False
    return True

def _reformat(distribution, transitions):
    """
    Reformat the input as dtype float numpy arrays.
    Also check for errors.
    @param distribution: initial state distribution numpy array
    @param transitions: transition matrix as numpy array
    """
    # get the initial state distribution as a numpy array
    np_distn = np.array(distribution, dtype=float)
    if not is_stochastic_vector(np_distn):
        msg = 'the initial distribution should be a stochastic vector'
        raise ValueError(msg)
    # get the transition matrix as a numpy array
    np_trans = np.array(transitions, dtype=float)
    if not is_right_stochastic_matrix(np_trans):
        msg = 'the transition matrix should be a right stochastic matrix'
        raise ValueError(msg)
    # the vector and matrix should be conformant
    nstates = np_distn.shape[0]
    if np_trans.shape != (nstates, nstates):
        msg_a = 'the number of states in the initial distribution does not '
        msg_b = 'match the number of states in the transition matrix'
        raise ValueError(msg_a + msg_b)
    return np_distn, np_trans

def forward(distribution, transitions,
        likelihoods_name, forward_name, scaling_name):
    """
    @param distribution: initial state distribution
    @param transitions: transition probabilities
    """
    np_distn, np_trans = _reformat(distribution, transitions)
    return hmmusbuf.forward(np_distn, np_trans,
        likelihoods_name, forward_name, scaling_name)

def backward(distribution, transitions,
        likelihoods_name, scaling_name, backward_name):
    """
    @param distribution: initial state distribution
    @param transitions: transition probabilities
    """
    np_distn, np_trans = _reformat(distribution, transitions)
    return hmmusbuf.backward(np_distn, np_trans,
        likelihoods_name, scaling_name, backward_name)

def posterior(distribution, transitions,
        forward_name, scaling_name, backward_name, posterior_name):
    """
    @param distribution: initial state distribution
    @param transitions: transition probabilities
    """
    np_distn, np_trans = _reformat(distribution, transitions)
    return hmmusbuf.posterior(np_distn, np_trans,
        forward_name, scaling_name, backward_name, posterior_name)

def fwdbwd_alldisk(distribution, transitions,
        likelihoods_name,
        forward_name, scaling_name, backward_name,
        posterior_name):
    """
    @param distribution: initial state distribution
    @param transitions: transition probabilities
    """
    forward(distribution, transitions,
            likelihoods_name, forward_name, scaling_name)
    backward(distribution, transitions,
            likelihoods_name, scaling_name, backward_name)
    posterior(distribution, transitions,
            forward_name, scaling_name, backward_name, posterior_name)

def fwdbwd_somedisk(distribution, transitions,
        likelihoods_name, posterior_name):
    """
    @param distribution: initial state distribution
    @param transitions: transition probabilities
    """
    np_distn, np_trans = _reformat(distribution, transitions)
    return hmmusbuf.fwdbwd_somedisk(np_distn, np_trans,
        likelihoods_name, posterior_name)

def fwdbwd_nodisk(distribution, transitions, np_likelihoods):
    """
    @param distribution: initial state distribution
    @param transitions: transition probabilities
    @param likelihoods: likelihoods at each state
    @return: posterior distribution at each state
    """
    np_distn, np_trans = _reformat(distribution, transitions)
    if len(np_likelihoods.shape) != 2:
        msg = 'the matrix of likelihoods should be rectangular'
        raise ValueError(msg)
    np_posterior = np.zeros_like(np_likelihoods)
    hmmusbuf.fwdbwd_nodisk(np_distn, np_trans, np_likelihoods, np_posterior)
    return np_posterior



"""
This is the setup script.
This script is automatically run by
easy_install or by pip on the user's machine when
she installs the module from pypi.

Here is some documentation for this process.
http://docs.python.org/extending/building.html

More info:
http://wiki.python.org/moin/Distutils/Tutorial

Register the metadata with pypi as follows:
python setup.py register

Send to pypi as follows:
python setup.py sdist upload --show-response
"""

from distutils.core import setup
from distutils.core import Extension
import os

from hmmus import hmmus_version


scripts = [
        'bin/hmm-od',
        'bin/hmm-regression-test',
        'bin/hmm-demo',
        'bin/hmm-forward',
        'bin/hmm-backward',
        'bin/hmm-posterior',
        'bin/hmm-fwdbwd-somedisk',
        'bin/hmm-fwdbwd-alldisk']

hmmusbuf = Extension(
        name='hmmusbuf',
        sources=['hmmusbufmodule.c', 'hmmguts/hmmguts.c'])

classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: C',
        'Programming Language :: Unix Shell',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules']

download_url_first = 'http://pypi.python.org/packages/source'
download_url_rest = 'h/hmmus/hmmus-' + hmmus_version + '.tar.gz'
download_url = os.path.join(download_url_first, download_url_rest)

# This seems like a standard method.
long_description = open('README.rst').read()

setup(
        name = 'hmmus',
        version = hmmus_version,
        author = 'Alex Griffing',
        author_email = 'argriffi@ncsu.edu',
        maintainer = 'Alex Griffing',
        maintainer_email = 'argriffi@ncsu.edu',
        url = 'http://github.com/argriffing/hmmus',
        download_url = download_url,
        description = 'Posterior decoding with a hidden Markov model',
        long_description = long_description,
        classifiers = classifiers,
        platforms = ['Linux'],
        license = 'MIT',
        ext_modules = [hmmusbuf],
        packages = ['hmmus'],
        scripts = scripts)
